package java.lang.reflect;

import sun.misc.ProxyGenerator;
import java.util.IdentityHashMap;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.BiFunction;
import java.util.Arrays;
import java.lang.ref.WeakReference;
import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.reflect.misc.ReflectUtil;
import sun.security.util.SecurityConstants;
import sun.misc.VM;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;
import java.util.Objects;
import java.io.Serializable;

public class Proxy implements Serializable
{
    private static final long serialVersionUID = -2222568056686623797L;
    private static final Class<?>[] constructorParams;
    private static final WeakCache<ClassLoader, Class<?>[], Class<?>> proxyClassCache;
    protected InvocationHandler h;
    private static final Object key0;
    
    private Proxy() {
    }
    
    protected Proxy(InvocationHandler invocationHandler) {
        Objects.requireNonNull(invocationHandler);
        this.h = invocationHandler;
    }
    
    @CallerSensitive
    public static Class<?> getProxyClass(ClassLoader classLoader, Class<?>... array) throws IllegalArgumentException {
        Class[] array2 = (Class<?>[])array.clone();
        if (System.getSecurityManager() != null) {
            checkProxyAccess(Reflection.getCallerClass(), classLoader, array2);
        }
        return getProxyClass0(classLoader, array2);
    }
    
    private static void checkProxyAccess(Class<?> clazz, ClassLoader classLoader, Class<?>... interfaces) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            ClassLoader classLoader2 = clazz.getClassLoader();
            if (VM.isSystemDomainLoader(classLoader) && !VM.isSystemDomainLoader(classLoader2)) {
                securityManager.checkPermission(SecurityConstants.GET_CLASSLOADER_PERMISSION);
            }
            ReflectUtil.checkProxyPackageAccess(classLoader2, interfaces);
        }
    }
    
    private static Class<?> getProxyClass0(ClassLoader classLoader, Class<?>... array) {
        if (array.length > 65535) {
            throw new IllegalArgumentException("interface limit exceeded");
        }
        return (Class)Proxy.proxyClassCache.get(classLoader, array);
    }
    
    @CallerSensitive
    public static Object newProxyInstance(ClassLoader classLoader, Class<?>[] array, InvocationHandler obj) throws IllegalArgumentException {
        Objects.requireNonNull(obj);
        Class[] array2 = (Class<?>[])array.clone();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            checkProxyAccess(Reflection.getCallerClass(), classLoader, array2);
        }
        Class<Object> proxyClass0 = getProxyClass0(classLoader, array2);
        try {
            if (securityManager != null) {
                checkNewProxyPermission(Reflection.getCallerClass(), proxyClass0);
            }
            final Constructor<Object> constructor = proxyClass0.getConstructor(Proxy.constructorParams);
            if (!Modifier.isPublic(proxyClass0.getModifiers())) {
                AccessController.doPrivileged(new PrivilegedAction<Void>() {
                    @Override
                    public Void run() {
                        constructor.setAccessible(true);
                        return null;
                    }
                });
            }
            return constructor.newInstance(new Object[] { obj });
        }
        catch (IllegalAccessException | InstantiationException cause) {
            throw new InternalError(cause.toString(), cause);
        }
        catch (InvocationTargetException ex) {
            Throwable cause2 = ex.getCause();
            if ((cause2 instanceof RuntimeException)) {
                throw (RuntimeException)cause2;
            }
            throw new InternalError(cause2.toString(), cause2);
        }
        catch (NoSuchMethodException cause3) {
            throw new InternalError(cause3.toString(), cause3);
        }
    }
    
    private static void checkNewProxyPermission(Class<?> clazz, Class<?> cls) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null && ReflectUtil.isNonPublicProxyClass(cls)) {
            ClassLoader classLoader = clazz.getClassLoader();
            ClassLoader classLoader2 = cls.getClassLoader();
            int lastIndex = cls.getName().lastIndexOf(46);
            String str = (lastIndex == -1) ? "" : cls.getName().substring(0, lastIndex);
            int lastIndex2 = clazz.getName().lastIndexOf(46);
            String anObject = (lastIndex2 == -1) ? "" : clazz.getName().substring(0, lastIndex2);
            if (classLoader2 != classLoader || !str.equals(anObject)) {
                securityManager.checkPermission(new ReflectPermission("newProxyInPackage." + str));
            }
        }
    }
    
    public static boolean isProxyClass(Class<?> clazz) {
        return Proxy.class.isAssignableFrom(clazz) && Proxy.proxyClassCache.containsValue(clazz);
    }
    
    @CallerSensitive
    public static InvocationHandler getInvocationHandler(Object o) throws IllegalArgumentException {
        if (!isProxyClass(o.getClass())) {
            throw new IllegalArgumentException("not a proxy instance");
        }
        InvocationHandler h = ((Proxy)o).h;
        if (System.getSecurityManager() != null) {
            Class<Object> class1 = h.getClass();
            if (ReflectUtil.needsPackageAccessCheck(Reflection.getCallerClass().getClassLoader(), class1.getClassLoader())) {
                ReflectUtil.checkPackageAccess(class1);
            }
        }
        return h;
    }
    
    private static native Class<?> defineClass0(ClassLoader p0, String p1, byte[] p2, int p3, int p4);
    
    static Class access$300(ClassLoader classLoader, String s, byte[] array, int n, int n2) {
        return defineClass0(classLoader, s, array, n, n2);
    }
    
    static {
        constructorParams = new Class[] { InvocationHandler.class };
        proxyClassCache = new WeakCache(new Proxy.KeyFactory(), new Proxy.ProxyClassFactory());
        key0 = new Object();
    }
    
    private static final class Key1 extends WeakReference<Class<?>>
    {
        private final int hash;
        
        Key1(Class<?> referent) {
            super(referent);
            this.hash = referent.hashCode();
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        @Override
        public boolean equals(Object o) {
            Class<Object> clazz;
            return this == o || (o != null && o.getClass() == Proxy.Key1.class && (clazz = (Class)this.get()) != null && clazz == ((Proxy.Key1)o).get());
        }
    }
    
    private static final class Key2 extends WeakReference<Class<?>>
    {
        private final int hash;
        private final WeakReference<Class<?>> ref2;
        
        Key2(Class<?> referent, Class<?> referent2) {
            super(referent);
            this.hash = 31 * referent.hashCode() + referent2.hashCode();
            this.ref2 = new WeakReference(referent2);
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        @Override
        public boolean equals(Object o) {
            Class<Object> clazz;
            Class<Object> clazz2;
            return this == o || (o != null && o.getClass() == Proxy.Key2.class && (clazz = (Class)this.get()) != null && clazz == ((Proxy.Key2)o).get() && (clazz2 = (Class)this.ref2.get()) != null && clazz2 == ((Proxy.Key2)o).ref2.get());
        }
    }
    
    private static final class KeyX
    {
        private final int hash;
        private final WeakReference<Class<?>>[] refs;
        
        KeyX(Class<?>[] a) {
            this.hash = Arrays.hashCode(a);
            this.refs = (WeakReference[])new WeakReference[a.length];
            for (int i = 0; i < a.length; ++i) {
                this.refs[i] = new WeakReference(a[i]);
            }
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        @Override
        public boolean equals(Object o) {
            return this == o || (o != null && o.getClass() == Proxy.KeyX.class && equals(this.refs, ((Proxy.KeyX)o).refs));
        }
        
        private static boolean equals(WeakReference<Class<?>>[] array, WeakReference<Class<?>>[] array2) {
            if (array.length != array2.length) {
                return false;
            }
            int i = 0;
            while (i < array.length) {
                Class<Object> clazz = (Class)array[i].get();
                if (clazz == null || clazz != array2[i].get()) {
                    return false;
                }
                ++i;
            }
            return true;
        }
    }
    
    private static final class KeyFactory implements BiFunction<ClassLoader, Class<?>[], Object>
    {
        @Override
        public Object apply(ClassLoader classLoader, Class<?>[] array) {
            switch (array.length) {
                case 1: {
                    return new Proxy.Key1(array[0]);
                }
                case 2: {
                    return new Proxy.Key2(array[0], array[1]);
                }
                case 0: {
                    return Proxy.key0;
                }
                default: {
                    return new KeyX(array);
                }
            }
        }
    }
    
    private static final class ProxyClassFactory implements BiFunction<ClassLoader, Class<?>[], Class<?>>
    {
        private static final String proxyClassNamePrefix = "$Proxy";
        private static final AtomicLong nextUniqueNumber;
        
        @Override
        public Class<?> apply(ClassLoader loader, Class<?>[] array) {
            IdentityHashMap identityHashMap = new IdentityHashMap(array.length);
            int length = array.length;
            int i = 0;
            while (i < length) {
                Class<?> obj = array[i];
                Class<Object> forName = null;
                try {
                    forName = Class.forName(obj.getName(), false, loader);
                }
                catch (ClassNotFoundException ex) {}
                if (forName != obj) {
                    throw new IllegalArgumentException(obj + " is not visible from class loader");
                }
                if (!forName.isInterface()) {
                    throw new IllegalArgumentException(forName.getName() + " is not an interface");
                }
                if (identityHashMap.put(forName, Boolean.TRUE) != null) {
                    throw new IllegalArgumentException("repeated interface: " + forName.getName());
                }
                ++i;
            }
            String s = null;
            int n = 17;
            for (Class<?> clazz : array) {
                if (!Modifier.isPublic(clazz.getModifiers())) {
                    n = 16;
                    String name = clazz.getName();
                    int lastIndex = name.lastIndexOf(46);
                    String s2 = (lastIndex == -1) ? "" : name.substring(0, lastIndex + 1);
                    if (s == null) {
                        s = s2;
                    }
                    else if (!s2.equals(s)) {
                        throw new IllegalArgumentException("non-public interfaces from different packages");
                    }
                }
            }
            if (s == null) {
                s = "com.sun.proxy.";
            }
            String string = s + "$Proxy" + Proxy.ProxyClassFactory.nextUniqueNumber.getAndIncrement();
            byte[] generateProxyClass = ProxyGenerator.generateProxyClass(string, array, n);
            try {
                return Proxy.access$300(loader, string, generateProxyClass, 0, generateProxyClass.length);
            }
            catch (ClassFormatError classFormatError) {
                throw new IllegalArgumentException(classFormatError.toString());
            }
        }
        
        static {
            nextUniqueNumber = new AtomicLong();
        }
    }
}
